
// RC circuit AC analysis
// Warning: not validated, unoptimized

// RC circuit:
//           _____
//     o----|_____|---o------o
//             R      |
//   Uin              |          Uout
//                   === C
//                    |
//     o--------------o------o
//

// Method: Runge-Kutta 4
// Needs linking with math library (GCC option -lm)
// Copyright: (c) Petr Peringer, FIT TU Brno, 2018

#include <stdio.h>
#include <math.h>               // sin(), M_PI

// System: RLC serial
#define SIZE 1                  // state vector size of the system
// model parameters:
      double R = 10;
const double C = 100e-6;
// input:
const double A  = 1.0;          // Amplitude (NOT effective)
      double freq = 50;         // frequency

inline double uin(double t) {
    return A*sin(2*M_PI*freq*t); // sin(\omega t)
}

inline double uout(double t, double y[]) {
    double u_C = y[0];
    return u_C;
}

// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    double u_C = y[0];
    double u_R = uin(t)-u_C;
    double i   = u_R/R;
    yin[0] = i/C;               // der(u_C)
}

// Runge-Kutta-4 step
void RK4step(double t, double h, double *y);

// Experiment control
void Run()
{
    // Simulation parameters:
    const double Fr = 1/(2*M_PI*R*C);   // frequency
    printf("# RC analysis\n");
    printf("# R=%g\n",R);
    printf("# C=%g\n",C);
    printf("# Fr=%g\n",Fr);
    double q=1.2;
    for(freq=Fr/50;freq<Fr*1000;freq*=q) {    // change input frequency
        q=(fabs(freq-Fr)/Fr < 0.1)?1.01:1.2;
        const double T    = 1/freq;     // period of input signal
        const double TEND = 1000*T;     // how many periods
        const double stepsize = T/50;   // RK4 step size
        double y[SIZE] = { 0, };        // initial values
        double umax=0;                  // amplitude
        double t = 0;                   // time
        while (t < TEND) {
            if(t>0.8*TEND && uout(t,y)>umax)
                 umax=uout(t,y);
            RK4step(t, stepsize, y);    // make step
            t += stepsize;              // advance the simulation time
        }
        if(umax<=0) printf("#");
        printf("%8.4g  %8g\n", freq, umax);
    }
    printf("\n");
}

int main()
{
    R = 10;
    Run();
    R = 1;
    Run();
    return 0;
}

// Runge-Kutta 4-th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ystart[SIZE];        // initial state
    // 4 stages results:
    double k1[SIZE];
    double k2[SIZE];
    double k3[SIZE];
    double k4[SIZE];

    int i;
    for (i = 0; i < SIZE; i++)  // save initial value
        ystart[i] = y[i];

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        k1[i] = h * yin[i];     // k1 = h * f(t, y(t))
        y[i] = ystart[i] + k1[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        k2[i] = h * yin[i];     // k2 = h * f(t+h/2, y(t)+k1/2)
        y[i] = ystart[i] + k2[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        k3[i] = h * yin[i];     // k3 = h * f(t+h/2, y(t)+k2/2)
        y[i] = ystart[i] + k3[i];
    }

    f(t + h, y, yin);           // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        k4[i] = h * yin[i];     // k4 = h * f(t+h, y(t)+k3)
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] = ystart[i] + k1[i] / 6 + k2[i] / 3 + k3[i] / 3 + k4[i] / 6;
    }
    // Return: y = new state at time t+h
}
